module TestingTypeUtils

using TypeUtils
using Test
using Base: OneTo

struct Foo{T1,T2}
    z::Complex{T1}
    r::T2
    i::Int
end

struct Bar{T}
    x::T
    y::Tuple{T,Int16,UInt8}
end

@testset "TypeUtils" begin
    @testset "parameterless()" begin
        @test parameterless(Array) === Array
        @test parameterless(AbstractVector) === AbstractArray
        @test parameterless(DenseMatrix) === DenseArray
        @test parameterless(Array{Float32}) === Array
        @test parameterless(DenseArray{Float32,3}) === DenseArray
        @test parameterless(Matrix{Float64}) === Array
        @test parameterless(typeof((1, 2, 3))) === Tuple
        @test parameterless(typeof((1, π, "e"))) === Tuple
    end

    @testset "as()" begin
        @test as(Int, 3) === 3
        @test as(Int, 1.0) === 1
        @test as(Int16, 1.0) === Int16(1)
        @test_throws InexactError as(Int, sqrt(2))
        @test_throws Exception as(Int, π)
        @test map(as(Int), (Int8(1), Int16(2), Int32(3), Int64(4))) === (1,2,3,4)

        @test as(Tuple, CartesianIndex((1,2,3))) === (1,2,3)
        @test as(NTuple{3}, CartesianIndex((1,2,3))) === (1,2,3)
        @test as(Tuple, CartesianIndices((2:5, -1:4))) === (2:5, -1:4)
        @test as(NTuple{2}, CartesianIndices((2:5, -1:4))) === (2:5, -1:4)

        @test as(CartesianIndex,()) === CartesianIndex()
        @test as(CartesianIndex, CartesianIndex((1,2,3))) === CartesianIndex((1,2,3))
        @test as(CartesianIndex{3}, CartesianIndex((1,2,3))) === CartesianIndex((1,2,3))
        @test as(CartesianIndex, (0x1,2,Int16(3))) === CartesianIndex((1,2,3))
        @test as(CartesianIndex{3}, (0x1,2,Int16(3))) === CartesianIndex((1,2,3))

        @test as(CartesianIndices,()) === CartesianIndices(())
        @test as(CartesianIndices,(2:3,6)) === CartesianIndices((2:3,6))
        @test as(CartesianIndices,CartesianIndices((2:3,6))) === CartesianIndices((2:3,6))
        @test as(CartesianIndices{3},(2:3,6,-1:4)) === CartesianIndices((2:3,6,-1:4))
        @test as(CartesianIndices{3},CartesianIndices((2:3,6,-1:4))) === CartesianIndices((2:3,6,-1:4))

        @test as(String, :hello) == "hello"
        @test as(String, :hello) isa String
        @test as(Symbol, "hello") === :hello
    end

    @testset "promote_eltype()" begin
        A = rand(Float32, 2, 3)
        B = ones(Int16, 2)
        C = zeros(Bool, 3, 4)
        @test promote_eltype() === promote_type()
        @test promote_eltype(A) === eltype(A)
        @test promote_eltype(B) === eltype(B)
        @test promote_eltype(C) === eltype(C)
        @test promote_eltype(A,B) === promote_type(eltype(A), eltype(B))
        @test promote_eltype(A,B,C) === promote_type(eltype(A), eltype(B), eltype(C))
    end

    @testset "convert_eltype()" begin
        # Abstract arrays.
        let A = rand(Float64, 2, 3), B = @inferred convert_eltype(Float32, A)
            @test convert_eltype(eltype(A), A) === A
            @test B isa AbstractArray{Float32,ndims(A)}
            @test B == Float32.(A)
        end
        # Base.OneTo
        let A = OneTo{Int32}(7), B = @inferred convert_eltype(Int16, A)
            @test convert_eltype(eltype(A), A) === A
            @test B isa OneTo{Int16}
            @test B == Int16.(A)
        end
        let A = OneTo(7), B = @inferred convert_eltype(Float32, A)
            @test convert_eltype(eltype(A), A) === A
            @test B isa AbstractRange{Float32}
            @test B == Float32.(A)
        end
        # UnitRange
        let A = 2:8, B = @inferred convert_eltype(Float32, A)
            @test convert_eltype(eltype(A), A) === A
            @test B isa AbstractRange{Float32}
            @test B == Float32.(A)
        end
        # OrdinalRange
        let A = 2.0:3.0:11.0, B = @inferred convert_eltype(Float32, A)
            @test convert_eltype(eltype(A), A) === A
            @test B isa AbstractRange{Float32}
            @test B == Float32.(A)
        end
        # LinRange
        let A = LinRange(-2.0, 3.0, 5), B = @inferred convert_eltype(Float32, A)
            @test convert_eltype(eltype(A), A) === A
            @test B isa LinRange{Float32}
            @test B == Float32.(A)
        end
        # Tuples.
        let A = (1, 2, 3) #= NTuple =#, B = @inferred convert_eltype(Float32, A)
            @test convert_eltype(eltype(A), A) === A
            @test B isa NTuple{<:Any,Float32}
            @test B == Float32.(A)
        end
        let A = (1, 2.0, 3) #= not NTuple but Tuple =#, B = @inferred convert_eltype(Float32, A)
            @test convert_eltype(eltype(A), A) === A
            @test B isa NTuple{<:Any,Float32}
            @test B == Float32.(A)
        end
    end

    @testset "as_eltype()" begin
        let A = rand(Float64, 3, 4, 5), B = @inferred as_eltype(Float32, A)
            @test B == Float32.(A)
            @test eltype(B) === Float32
            @test length(B) === length(A)
            @test size(B) === size(A)
            @test axes(B) === axes(A)
            A[1,2,3] = -7
            @test B[1,2,3] === Float32(-7)
            B[1,2,3] = 19
            @test A[1,2,3] == 19
        end
        let A = view(rand(Float64, 3, 4, 5), :, 2, :), B = @inferred as_eltype(Float32, A)
            @test as_eltype(eltype(A), A) === A
            @test B == Float32.(A)
            @test eltype(B) === Float32
            @test length(B) === length(A)
            @test size(B) === size(A)
            @test axes(B) === axes(A)
            A[2,3] = -7
            @test B[2,3] === Float32(-7)
            B[2,3] = 19
            @test A[2,3] == 19
        end
        let A = 1:5
            @test as_eltype(eltype(A), A) === A
            @test as_eltype(Float32, A) == Float32.(A)
        end
        let A = 2.0:3.0:11.0
            @test as_eltype(eltype(A), A) === A
            @test as_eltype(Float32, A) == Float32.(A)
        end
    end

    @testset "return_type()" begin
        @test return_type(cos, Float32) === typeof(cos(zero(Float32)))
        @test return_type(atan, Int32) === typeof(atan(zero(Int32)))
        @test return_type(atan, Int32, Int16) === typeof(atan(zero(Int32), one(Int16)))
    end

    @testset "as_return()" begin
        f = @inferred as_return(Float32, atan)
        g = @inferred as_return(BigFloat, f)
        h = @inferred as_return(return_type(f), f)
        @test g isa parameterless(typeof(f))
        @test h === f
        @test parent(f) === atan
        @test parent(g) === parent(f)
        @test return_type(f) === Float32
        @test return_type(g) === BigFloat
        @test return_type(typeof(f)) === Float32
        @test return_type(typeof(g)) === BigFloat
        @test Base.return_types(f) === (return_type(f),)
        @test Base.return_types(g) === (return_type(g),)
        @test Base.promote_op(f, Int8) === return_type(f)
        @test Base.promote_op(f, Int8, Int16) === return_type(f)
        @test Base.promote_op(g, Float32) === return_type(g)
        @test f(0) isa Float32
        @test f(-0.3) isa Float32
        @test g(0, 1) isa BigFloat
        @test g(-0.3) isa BigFloat
    end

    @testset "Destructure and restructure" begin
        # Test with structures.
        obj = Foo(2.0f0 - 3.0f0im, π, 42) # NOTE π is special
        @test obj isa Foo{Float32,typeof(π)}
        @test isconcretetype(typeof(obj))
        @test struct_length(obj) == 4
        vals = @inferred destructure(obj)
        @test vals === (2.0f0, -3.0f0, π, 42)
        @test obj === @inferred restructure(typeof(obj), vals)
        @test obj === restructure(parameterless(typeof(obj)), vals)
        @test obj === @inferred restructure(typeof(obj), (0, 1, vals...); offset=2)
        vec = Vector{Float64}(undef, 6)
        @test destructure!(vec, obj)[1:4] ≈ collect(vals)
        @test destructure!(vec, obj; offset=1)[2:5] ≈ collect(vals)
        x = restructure(Foo{Float64,Float64}, vec; offset=1)
        @test x isa Foo{Float64,Float64}
        @test x.z ≈ obj.z
        @test x.r ≈ obj.r
        @test x.i == obj.i
        # Test with tuples.
        a = (1, 2, (3, (4, 5)), π, ())
        @test struct_length(a) == 7
        let vals = @inferred destructure(a)
            if VERSION < v"1.0" || v"1.5" < VERSION < v"1.10"
                # FIXME: not all versions of Julia corectly infer the result.
                @test a === restructure(typeof(a), vals)
            else
                @test a === @inferred restructure(typeof(a), vals)
            end
        end
        # Test with structure and tuples.
        b = (1, 2, obj, (3, 4))
        @test struct_length(b) == 4 + struct_length(obj)
        let vals = @inferred destructure(b)
            if VERSION < v"1.0" || v"1.5" < VERSION < v"1.10"
                # FIXME: not all versions of Julia corectly infer the result.
                @test b === restructure(typeof(b), vals)
            else
                @test b === @inferred restructure(typeof(b), vals)
            end
        end
        c = Bar{Float32}(1, (2, 3, 4))
        @test struct_length(c) == 4
        let vals = @inferred destructure(c)
            @test c === @inferred restructure(typeof(c), vals)
        end
    end
end

end # module
